from .exception import InterfaceException
from .serial import SerialInterfaceException, SerialInterface
import sys

if sys.platform == "linux":
    from .scsi_linux import LinuxSCSIInterface as SCSIInterface
    from .scsi_linux import LinuxSCSIInterfaceException as SCSIInterfaceException
elif sys.platform == "win32":
    from .scsi_windows import WindowsSCSIInterface as SCSIInterface
    from .scsi_windows import WindowsSCSIInterfaceException as SCSIInterfaceException
else:
    raise NotImplemented("%s not supported by this library" % (sys.platform,))



import serial.win32 as win32
from ctypes.wintypes import UINT, ULONG, USHORT, DWORD, BOOL, HANDLE
from ctypes.wintypes import LPCWSTR
import ctypes

# an interface for communicating with reader over SCSI, implemented in the style of pyserial

DRIVE_REMOVABLE = 2
DRIVE_CDROM = 5

FILE_SHARE_READ = 1
FILE_SHARE_WRITE = 2

UCHAR = ctypes.c_ubyte
PVOID = win32.PVOID
LPOVERLAPPED = win32.LPOVERLAPPED
LPVOID = win32.LPVOID
LPDWORD = win32.LPDWORD

try:
    GetDriveTypeW = win32._stdcall_libraries["kernel32"].GetDriveTypeW
    GetDriveTypeW.restype = UINT
    GetDriveTypeW.argtypes = [
        LPCWSTR,
    ]
    GetDriveType = GetDriveTypeW  # alias
except AttributeError:
    from ctypes.wintypes import LPCSTR

    GetDriveTypeA = win32._stdcall_libraries["kernel32"].GetDriveTypeA
    GetDriveTypeA.restype = UINT
    GetDriveTypeA.argtypes = [
        LPCSTR,
    ]
    GetDriveType = GetDriveTypeA  # alias

DeviceIoControl = win32._stdcall_libraries["kernel32"].DeviceIoControl
DeviceIoControl.restype = BOOL
DeviceIoControl.argtypes = [
    HANDLE,
    DWORD,
    LPVOID,
    DWORD,
    LPVOID,
    DWORD,
    LPDWORD,
    LPOVERLAPPED,
]


class SCSI_PASS_THROUGH_DIRECT(ctypes.Structure):
    pass


SCSI_PASS_THROUGH_DIRECT._fields_ = [
    ("Length", USHORT),
    ("ScsiStatus", UCHAR),
    ("PathId", UCHAR),
    ("TargetId", UCHAR),
    ("Lun", UCHAR),
    ("CdbLength", UCHAR),
    ("SenseInfoLength", UCHAR),
    ("DataIn", UCHAR),
    ("DataTransferLength", ULONG),
    ("TimeOutValue", ULONG),
    ("DataBuffer", PVOID),
    ("SenseInfoOffset", ULONG),
    ("Cdb", UCHAR * 16),
]


class SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER(ctypes.Structure):
    pass


SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER._fields_ = [
    ("sptd", SCSI_PASS_THROUGH_DIRECT),
    ("Filler", ULONG),
    ("ucSenseBuf", UCHAR * 32),
]


def CTL_CODE(DeviceType, Function, Method, Access):
    return (DeviceType << 16) | (Access << 14) | (Function << 2) | Method


FILE_DEVICE_CONTROLLER = 0x00000004
IOCTL_SCSI_BASE = FILE_DEVICE_CONTROLLER
CDB10GENERIC_LENGTH = 10
SCSI_IOCTL_DATA_OUT = 0
SCSI_IOCTL_DATA_IN = 1
SCSI_IOCTL_DATA_UNSPECIFIED = 2

METHOD_BUFFERED = 0
FILE_READ_ACCESS = 1
FILE_WRITE_ACCESS = 2
IOCTL_SCSI_PASS_THROUGH_DIRECT = CTL_CODE(
    IOCTL_SCSI_BASE, 0x0405, METHOD_BUFFERED, FILE_READ_ACCESS | FILE_WRITE_ACCESS
)


class WindowsSCSIInterfaceException(Exception):
    pass


class WindowsSCSIInterface(object):
    def __init__(self, port: str):
        self._drive = port

        self._open()

    def _open(self):
        dos_drive_path = self._drive
        ddk_drive_path = "\\\\.\\%s" % (self._drive,)

        # Check drive type
        drive_type = GetDriveType(dos_drive_path)
        if drive_type not in (DRIVE_REMOVABLE, DRIVE_CDROM):
            raise WindowsSCSIInterfaceException("Drive type is incorrect.")

        self._port_handle = win32.CreateFile(
            ddk_drive_path,
            win32.GENERIC_WRITE | win32.GENERIC_READ,
            FILE_SHARE_READ | FILE_SHARE_WRITE,
            None,
            win32.OPEN_EXISTING,
            win32.FILE_ATTRIBUTE_NORMAL,  # | win32.FILE_FLAG_OVERLAPPED,
            0,
        )

        if self._port_handle == win32.INVALID_HANDLE_VALUE:
            self._port_handle = None
            raise WindowsSCSIInterfaceException("Could not open drive.")

    def _scsi_operation(self, pbuf, size, read, timeout=10):
        sptdwb = SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER()
        ctypes.memset(ctypes.addressof(sptdwb), 0, ctypes.sizeof(sptdwb))

        sptdwb.sptd.Length = ctypes.sizeof(SCSI_PASS_THROUGH_DIRECT)
        sptdwb.sptd.PathId = 0
        sptdwb.sptd.TargetId = 1
        sptdwb.sptd.Lun = 0
        sptdwb.sptd.CdbLength = CDB10GENERIC_LENGTH
        sptdwb.sptd.SenseInfoLength = 0
        sptdwb.sptd.DataIn = SCSI_IOCTL_DATA_IN if read else SCSI_IOCTL_DATA_OUT
        sptdwb.sptd.DataTransferLength = size
        sptdwb.sptd.TimeOutValue = timeout
        sptdwb.sptd.DataBuffer = pbuf
        sptdwb.sptd.SenseInfoOffset = (
            SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER.ucSenseBuf.offset
        )
        sptdwb.sptd.Cdb[0] = 0xEF
        sptdwb.sptd.Cdb[1] = 0xFF if read else 0xFE

        length = DWORD(ctypes.sizeof(SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER))
        returned = DWORD(0)

        result_ok = DeviceIoControl(
            self._port_handle,
            IOCTL_SCSI_PASS_THROUGH_DIRECT,
            ctypes.pointer(sptdwb),
            length,
            ctypes.pointer(sptdwb),
            length,
            ctypes.byref(returned),
            None,
        )

        if not result_ok and win32.GetLastError() not in (
            win32.ERROR_SUCCESS,
            win32.ERROR_IO_PENDING,
        ):
            raise WindowsSCSIInterfaceException(
                "DeviceIoControl failed ({!r})".format(ctypes.WinError())
            )

        if returned.value < (
            SCSI_PASS_THROUGH_DIRECT.ScsiStatus.offset
            + SCSI_PASS_THROUGH_DIRECT.ScsiStatus.size
        ):
            raise WindowsSCSIInterfaceException(
                "Not enough SCSI information returned to determine error"
            )

        if sptdwb.sptd.ScsiStatus != 0:
            raise WindowsSCSIInterfaceException(
                "SCSI Operation returned %d" % (sptdwb.sptd.ScsiStatus,)
            )

        return sptdwb.sptd.DataTransferLength

    def read(self, size=1):
        assert size != 0
        # Allocate data
        buf = ctypes.create_string_buffer(size)
        data_read = self._scsi_operation(ctypes.addressof(buf), size, True)

        assert data_read == size
        return buf.raw[:data_read]

    def write(self, data):
        assert len(data) != 0
        buf = ctypes.create_string_buffer(len(data))
        buf.raw = data

        return self._scsi_operation(ctypes.addressof(buf), len(buf), False)

    def close(self):
        win32.CloseHandle(self._port_handle)



import logging
import serial
import time

from .exception import InterfaceException

logger = logging.getLogger(__name__)


class SerialInterfaceException(InterfaceException):
    pass


class SerialInterface(object):
    _DEFAULT_BAUD_RATE = 9600
    _DEFAULT_BYTESIZE = serial.EIGHTBITS
    _DEFAULT_TIMEOUT = 2  # seconds
    _BUFFERED_DELAY = 0.05
    _FRAGMENT_SIZE = 512

    def __init__(
        self,
        port,
        baudrate=_DEFAULT_BAUD_RATE,
        bytesize=_DEFAULT_BYTESIZE,
        timeout=_DEFAULT_TIMEOUT,
    ):
        self._port = port
        try:
            self._serial = serial.Serial(
                port=self._port, baudrate=baudrate, bytesize=bytesize, timeout=timeout
            )
        except serial.SerialException as e:
            logger.error("Could not open the serial device: %s" % (e,))
            raise SerialInterfaceException(e)

        if self._serial.is_open:
            self._serial.close()

        self._serial.open()
        self._serial.reset_output_buffer()
        self._serial.reset_input_buffer()

    def _flush(self):
        while len(self._serial.read(self._serial.in_waiting)) > 0:
            self._delay(self._BUFFERED_DELAY)

    def _buffered_read(self, count):
        data = bytes()
        fragment = self._serial.read(self._serial.in_waiting)
        while len(data) < count:
            self._delay(self._BUFFERED_DELAY)
            logger.debug("Read fragment of %d size" % (len(fragment),))
            data += fragment
            fragment = self._serial.read(self._serial.in_waiting)

        assert len(data) == count
        return data

    def _delay(self, seconds):
        time.sleep(seconds)

    def write(self, data):
        return self._serial.write(data)

    def read(self, to_read):
        if to_read > SerialInterface._FRAGMENT_SIZE:
            return self._buffered_read(to_read)
        return self._serial.read(to_read)

    def close(self):
        return self._serial.close()



from . import packets
from .interfaces import SCSIInterface, SerialInterface, InterfaceException

import logging
import contextlib
import threading
import time
import os
import PIL
import PIL.Image
import functools

from typing import ContextManager, Optional, Callable, TypeVar, Tuple, ClassVar, Union

logger = logging.getLogger(__name__)

RT = TypeVar("RT")


def retry_three_times(func: Callable[..., RT]) -> Callable[..., RT]:
    def wrapper(*args, **kwargs) -> RT:
        for _ in range(3):
            if func(*args, **kwargs):
                return True

    return wrapper


def save_bitmap_to_file(path: str, bitmap: bytes) -> None:
    img = PIL.Image.frombytes("L", (202, 258), bitmap, "raw")
    img.save(path, "BMP")


class GT521F32Exception(Exception):
    pass


class GT521F32(object):
    _PROMPT_INTERVAL: ClassVar[int] = 0.1
    _port: str
    _interface: Union[SerialInterface, SCSIInterface]
    _firmware_version: Optional[str]
    _iso_area_max_size: Optional[int]
    _device_serial_number: Optional[str]
    _cancel: threading.Event

    @staticmethod
    def _choose_interface_type(port):
        if any(
            port.startswith(_) for _ in ("COM", "/dev/tty")
        ):  # Any other path patterns?
            return SerialInterface
        elif port.startswith("/dev/sg") or (
            port[0].isalpha() and port[1] == ":" and len(port) == 2
        ):
            return SCSIInterface
        raise GT521F32Exception("Could not derive interface type from port path")

    def __init__(self, port: str, baudrate: Optional[None] = None):
        self._port = port
        try:
            interface_cls = GT521F32._choose_interface_type(port)
            logger.debug("Chose interface type %s" % (interface_cls.__name__,))
            if baudrate is not None:
                if interface_cls is not SerialInterface:
                    raise GT521F32Exception(
                        "Baud rate can only be given for serial interfaces."
                    )
                self._interface = interface_cls(port=port, baudrate=baudrate)
            else:
                self._interface = interface_cls(port=port)
        except InterfaceException as e:
            logger.error("Could not open the fingerprint device: %s" % (e,))
            raise GT521F32Exception("Failed to open the fingerprint device.")

        self._firmware_version = None
        self._iso_area_max_size = None
        self._device_serial_number = None

        self._cancel = threading.Event()

    def _delay(self, seconds: int) -> None:
        time.sleep(seconds)

    def send_command(self, command: str, parameter: int) -> Tuple[int, int]:
        if command not in packets.command_codes.keys():
            logger.error("Bad command.")
            raise GT521F32Exception("Invalid command.")

        command_code = packets.command_codes[command]
        command_packet = packets.CommandPacket(
            parameter=parameter, command=command_code
        )

        self._interface.write(command_packet.to_bytes())

        # read response
        to_read = packets.ResponsePacket().byte_size()
        response_bytes = self._interface.read(to_read)

        response_packet, _ = packets.ResponsePacket.from_bytes(response_bytes)
        if response_packet is None:
            logger.error("Command failed.")
            raise GT521F32Exception("Command failed.")

        if not response_packet.ok:
            logger.debug(
                "Command responded with code %x and error %04x"
                % (response_packet.response_code, response_packet.parameter)
            )

        return response_packet.response_code, response_packet.parameter

    def usb_internal_check(self) -> True:
        _, _ = self.send_command("USB_INTERNAL_CHECK", 0)

    def change_baud_rate(self, baudrate: int) -> None:
        # Not really relevant for USB (scsi) mode, but the command
        # is still supported
        response_code, parameter = self.send_command("CHANGE_BAUDRATE", baudrate)
        if response_code != packets.ACK_OK:
            logger.error(
                "ChangeBaudRate error: %s"
                % (packets.reverse(packets.response_error)[parameter],)
            )

    def change_baud_rate_and_reopen(self, baudrate: int) -> None:
        # We can send the command and it wont do any harm, but we dont want the
        # interface to be reopened, so unless we are already using a serial interface, do not proceed
        if type(self._interface) is not SerialInterface:
            raise NotImplemented(
                "Baud-rate not supported for interface type %s"
                % (type(self._interface),)
            )
        self.change_baud_rate(baudrate)
        self._interface.close()
        self._interface = SerialInterface(port=self._port, baudrate=baudrate)

    @property
    def firmware_version(self):
        return self._firmware_version

    @property
    def iso_area_max_size(self):
        return self._iso_area_max_size

    @property
    def device_serial_number(self):
        return self._device_serial_number

    def open(self) -> Tuple[str, int, str]:
        _, _ = self.send_command("OPEN", 1)

        # read data response
        to_read = packets.OpenDataPacket().byte_size()
        response_bytes = self._interface.read(to_read)

        open_data_response, _ = packets.OpenDataPacket.from_bytes(response_bytes)
        self._firmware_version, self._iso_area_max_size, self._device_serial_number = (
            open_data_response.firmware_version,
            open_data_response.iso_area_max_size,
            open_data_response.device_serial_number,
        )

        logger.info("Firmware version: %s" % (open_data_response.firmware_version,))
        logger.info("Iso area max size: %s" % (open_data_response.iso_area_max_size,))
        logger.info("Serial number: %s" % (open_data_response.device_serial_number,))

        return (
            self.firmware_version,
            self.iso_area_max_size,
            self.device_serial_number,
        )

    def module_info(self) -> Tuple[str, str, int, int, int, int, int, int]:
        response_code, parameter = self.send_command("MODULE_INFO", 0)

        # read data response
        to_read = parameter + packets.DataPacket().byte_size()
        response_bytes = self._interface.read(to_read)

        module_info_known_size = packets.ModuleInfoDataPacket().byte_size()
        if to_read > module_info_known_size:
            logger.error("Module info returned more bytes than expected.")

        module_info_packet, _ = packets.ModuleInfoDataPacket.from_bytes(response_bytes)

        logger.info("Sensor: %s" % (module_info_packet.sensor,))
        logger.info("Engine Version: %s" % (module_info_packet.engine_version,))
        logger.info("Raw Image Width: %s" % (module_info_packet.raw_img_width,))
        logger.info("Raw Image Height: %s" % (module_info_packet.raw_img_height,))
        logger.info("Image Height: %s" % (module_info_packet.img_width,))
        logger.info("Image Width: %s" % (module_info_packet.img_height,))
        logger.info("Max record count: %s" % (module_info_packet.max_record_count,))
        logger.info("Enroll count: %s" % (module_info_packet.enroll_count,))
        logger.info("Template size: %s" % (module_info_packet.template_size,))

        return (
            module_info_packet.sensor,
            module_info_packet.engine_version,
            module_info_packet.raw_img_width,
            module_info_packet.raw_img_height,
            module_info_packet.img_width,
            module_info_packet.img_height,
            module_info_packet.max_record_count,
            module_info_packet.enroll_count,
            module_info_packet.template_size,
        )

    def __delete__(self) -> None:
        self.close()

    def close(self) -> None:
        if False:
            # does nothing
            self.send_command("CLOSE", 0)
        self.change_baud_rate(9600)
        self._interface.close()

    def enroll_start(self, user_id: int) -> bool:
        response_code, parameter = self.send_command("ENROLL_START", user_id)
        if response_code != packets.ACK_OK:
            logger.error(
                "EnrollStart error: %s"
                % (packets.reverse(packets.response_error)[parameter],)
            )
            return False
        return True

    @retry_three_times
    def enroll_n(self, n: int, save_enroll_photos: bool = False) -> bool:
        self.prompt_finger_and_capture()

        if save_enroll_photos:
            # Save image before proceeding
            out_path = "Enroll%d.bmp" % (n,)
            logger.info("Saving Enroll%d to %s" % (n, out_path))
            save_bitmap_to_file(out_path, self.get_image())

        response_code, parameter = self.send_command("ENROLL%d" % (n,), 0)
        if response_code != packets.ACK_OK:
            error_code = packets.reverse(packets.response_error).get(parameter, None)
            if error_code is None:
                logger.error(
                    "Enroll%d error: %s" % (n, "Duplicate ID: %d" % (parameter,))
                )
                return True  # fast fail

            logger.error("Enroll%d error: %s" % (n, error_code))
            return False  # Will lead to retry

        logger.debug("Enroll%d succeeded." % (n,))
        return True

    def enroll_user(self, user_id: int, save_enroll_photos: bool = False) -> bool:
        attempts = 0
        if not self.enroll_start(user_id):
            return False

        for i in range(1, 4):
            with self.prompt_finger():
                if not self.enroll_n(
                    i, save_enroll_photos
                ):  # Not sure why this only works when reentering
                    logger.debug(
                        "Enrollment for user id %d failed, aborting." % (user_id,)
                    )
                    break
        else:
            logger.debug("Enroll user id: %d succeeded." % (user_id,))
            return True

        return False

    def identify(self) -> Optional[int]:
        self.prompt_finger_and_capture()

        response_code, parameter = self.send_command("IDENTIFY", 0)
        if response_code != packets.ACK_OK:
            logger.error(
                "Identify error: %s"
                % (packets.reverse(packets.response_error)[parameter],)
            )
            return None

        return parameter

    def get_raw_image_safe(self) -> Optional[bytes]:
        with self.led():  # Undocumented, but sensor crashes if led is off
            return self._get_raw_image()

    def _get_raw_image(self) -> Optional[bytes]:
        # Do not call this with the led off
        response_code, parameter = self.send_command("GET_RAWIMAGE", 0)
        if response_code != packets.ACK_OK:
            logger.error(
                "GetRawImage error: %s"
                % (packets.reverse(packets.response_error)[parameter],)
            )
            return

        # read data response
        logger.info("Downloading raw image...")
        to_read = packets.GetRawImageDataPacket().byte_size()
        response_bytes = self._interface.read(to_read)

        get_raw_image_data_response, _ = packets.GetRawImageDataPacket.from_bytes(
            response_bytes
        )

        return get_raw_image_data_response.raw_bitmap

    def get_image(self) -> Optional[bytes]:
        response_code, parameter = self.send_command("GET_IMAGE", 0)
        if response_code != packets.ACK_OK:
            logger.error(
                "GetImage error: %s"
                % (packets.reverse(packets.response_error)[parameter],)
            )
            return

        # read data response
        logger.info("Downloading image...")
        to_read = packets.GetImageDataPacket().byte_size()
        response_bytes = self._interface.read(to_read)

        get_image_data_response, _ = packets.GetImageDataPacket.from_bytes(
            response_bytes
        )

        return get_image_data_response.bitmap

    @contextlib.contextmanager
    def led(self) -> ContextManager[None]:
        self.set_led(True)
        yield
        self.set_led(False)

    def set_led(self, onoff: bool) -> None:
        assert type(onoff) is bool
        # Cannot fail
        _, _ = self.send_command("CMOS_LED", int(onoff))

    def capture(self, best_image: bool = False) -> bool:
        assert type(best_image) is bool
        response_code, parameter = self.send_command("CAPTURE", int(best_image))
        if response_code != packets.ACK_OK:
            logger.error(
                "Capture error: %s"
                % (packets.reverse(packets.response_error)[parameter],)
            )
            return False

        return True

    def get_enrolled_count(self) -> None:
        # Supposedly this cannot fail?
        _, parameter = self.send_command("ENROLL_COUNT", 0)
        return parameter

    def is_id_enrolled(self, user_id: int) -> bool:
        response_code, parameter = self.send_command("CHECK_ENROLLED", user_id)
        if response_code != packets.ACK_OK:
            logger.error(
                "CheckEnroll %d error: %s"
                % (user_id, packets.reverse(packets.response_error)[parameter])
            )
            return False
        return True

    def delete_id(self, user_id: int) -> bool:
        response_code, parameter = self.send_command("DELETE_ID", user_id)
        if response_code != packets.ACK_OK:
            logger.error(
                "DeleteID %d error: %s"
                % (user_id, packets.reverse(packets.response_error)[parameter])
            )
            return False

        return True

    def delete_all(self) -> bool:
        response_code, parameter = self.send_command("DELETE_ALL", 0)
        if response_code != packets.ACK_OK:
            logger.error(
                "DeleteAll error: %s"
                % (packets.reverse(packets.response_error)[parameter],)
            )
            return False

        return True

    def verify(self, user_id: int) -> bool:
        self.prompt_finger_and_capture()

        response_code, parameter = self.send_command("VERIFY", user_id)
        if response_code != packets.ACK_OK:
            logger.error(
                "Verify %d error: %s"
                % (user_id, packets.reverse(packets.response_error)[parameter])
            )
            return False

        return True

    def save_image_to_bmp(self, path: str) -> None:
        self.prompt_finger_and_capture()
        save_bitmap_to_file(path, self.get_image())

    # Utitilies
    def is_finger_pressed(self) -> bool:
        response_code, parameter = self.send_command("IS_PRESS_FINGER", 0)
        if response_code != packets.ACK_OK:
            logger.error(
                "Verify %d error: %s"
                % (user_id, packets.reverse(packets.response_error)[parameter])
            )
            return False
        return not bool(parameter)

    def cancel(self) -> None:
        self._cancel.set()

    def wait_for_finger_press(self, interval: float = _PROMPT_INTERVAL) -> None:
        while not self._cancel.is_set() and not self.is_finger_pressed():
            self._delay(interval)

        if self._cancel.is_set():
            logger.info("Cancelled action.")
            self._cancel.clear()

    def prompt_finger_and_capture(self) -> None:
        with self.prompt_finger():
            self.capture()

    @contextlib.contextmanager
    def prompt_finger(self) -> ContextManager[None]:
        with self.led():
            self.wait_for_finger_press(self._PROMPT_INTERVAL)
            yield



import signal
import sys
import time
from PyQt5.QtCore import QThread, Qt, pyqtSignal, pyqtSlot
from PyQt5.QtWidgets import QApplication, QWidget, QLabel
from PyQt5.QtGui import QIcon, QPixmap, QImage

import threading
import PIL
import PIL.Image
from PIL.ImageQt import ImageQt

import GT521F32


class ViewerThread(QThread):
    set_image = pyqtSignal(QImage)
    _cancel: threading.Event

    def __init__(self, parent, reader):
        super().__init__(parent)
        self._reader = reader
        self._cancel = threading.Event()

    def run(self):
        with self._reader.led():
            while not self._cancel.is_set():
                time.sleep(1.0 / 25)  # 25 fps
                data = self._reader._get_raw_image()
                if data:
                    img = PIL.Image.frombytes("L", (160, 120), data, "raw")
                    self.set_image.emit(ImageQt(img))


class Viewer(QWidget):
    def __init__(self, reader):
        super().__init__()
        self._reader = reader
        self.left = 10
        self.top = 10
        self.width = 640
        self.height = 480
        self.initUI()

    def initUI(self):
        self.setWindowTitle("GT521F32")
        self.setGeometry(self.left, self.top, self.width, self.height)

        self._thread = ViewerThread(self, self._reader)
        self._thread.set_image.connect(self.set_image)

        self.label = QLabel(self)
        self.label.resize(160 * 3, 120 * 3)
        self.resize(160 * 3, 120 * 3)
        self._thread.start()
        self.show()

    @pyqtSlot(QImage)
    def set_image(self, img):
        self.label.setPixmap(QPixmap.fromImage(img).scaled(160 * 3, 120 * 3))

    def closeEvent(self, event):
        self._thread._cancel.set()
        self._thread.wait()
        event.accept()


def main():
    signal.signal(signal.SIGINT, signal.SIG_DFL)  # Exit on Ctrl C
    app = QApplication(sys.argv)
    ex = Viewer(GT521F32.GT521F32(sys.argv[1]))
    ex.show()
    sys.exit(app.exec_())



import sys
import time

import argparse
import PIL.ImageTk
import PIL.Image
import tkinter
import threading

import GT521F32

from typing import Tuple, ClassVar


class GT521F32Viewer(object):
    _FRAME_RATE: ClassVar[int] = 24
    _DIMENSIONS: Tuple[int, int] = (
        160,
        120,
    )  # Might be better to retrieve this from reader

    _root: tkinter.Tk
    _image_panel: tkinter.Label
    _reader: GT521F32.GT521F32
    _scale_factor: int
    _stop: bool

    def __init__(self, reader: GT521F32.GT521F32, scale_factor: int = 1):
        self._reader = reader
        self._scale_factor = scale_factor

        self._stop = False

        self._root = tkinter.Tk()
        self._image_panel = tkinter.Label(self._root)
        self._image_panel.pack(padx=0, pady=0)

        self._root.title("GT521F32")
        self._root.geometry(
            "%dx%d" % tuple(_ * scale_factor for _ in GT521F32Viewer._DIMENSIONS)
        )
        self._root.resizable(0, 0)
        self._root.wm_protocol("WM_DELETE_WINDOW", self.stop)

    def _video_loop(self):
        data = self._reader._get_raw_image()
        if data:
            image = PIL.Image.frombytes("L", self._DIMENSIONS, data, "raw")
            self._update(image)

        if not self._stop:
            self._root.after(1_000 // GT521F32Viewer._FRAME_RATE, self._video_loop)

    def start(self):
        self._reader.set_led(True)
        self._root.after(1_000 // GT521F32Viewer._FRAME_RATE, self._video_loop)
        self._root.mainloop()

    def _update(self, image: PIL.ImageTk.PhotoImage):
        if self._scale_factor != 1:
            image = image.resize(
                (
                    int(image.width * self._scale_factor),
                    int(image.height * self._scale_factor),
                )
            )
        # To avoid undue GC
        self._image_tk = PIL.ImageTk.PhotoImage(image)
        self._image_panel.config(image=self._image_tk)

    def stop(self):
        self._stop = True
        self._reader.set_led(False)
        self._root.quit()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--device", required=True, help="Path to GT521F32 device")
    parser.add_argument(
        "-f", "--scale_factor", type=int, default=1.5, help="Image scaling factor."
    )
    args = parser.parse_args()

    try:
        v = GT521F32Viewer(GT521F32.GT521F32(args.device), args.scale_factor)
        v.start()
    except GT521F32.GT521F32Exception:
        print("Could not open fingerprint device.")
    except (KeyboardInterrupt, InterruptedError):
        v.stop()



import sys
import time
from PyQt5.QtCore import QThread, Qt, pyqtSignal, pyqtSlot
from PyQt5.QtWidgets import QApplication, QWidget, QLabel
from PyQt5.QtGui import QIcon, QPixmap, QImage

import PIL
import PIL.Image

import GT521F32


class ViewerThread(QThread):
    set_image = pyqtSignal(QImage)

    def __init__(self, parent, reader):
        super().__init__(parent)
        self._reader = reader

    def run(self):
        with self._reader.led():
            while True:
                self._reader.capture()
                data = self._reader.get_image()
                img = QImage(data, 202, 258, QImage.Format_RGB888)
                self.set_image.emit(img)
                time.sleep(0.5)
                return


class Viewer(QWidget):
    def __init__(self, reader):
        super().__init__()
        self._reader = reader
        self.left = 10
        self.top = 10
        self.width = 640
        self.height = 480
        self.initUI()

    def initUI(self):
        self.setWindowTitle("GT521F32")
        self.setGeometry(self.left, self.top, self.width, self.height)

        self._thread = ViewerThread(self, self._reader)
        self._thread.set_image.connect(self.set_image)

        self.label = QLabel(self)
        self.resize(202, 258)
        self._thread.start()
        self.show()

    @pyqtSlot(QImage)
    def set_image(self, img):
        print("yooo")
        self.label.setPixmap(QPixmap.fromImage(img))


if __name__ == "__main__":
    app = QApplication(sys.argv)
    ex = Viewer(GT521F32.GT521F32("/dev/sg2"))
    ex.show()
    sys.exit(app.exec_())



import logging
import serial
import sgio
import time

from .exception import InterfaceException

from typing import BinaryIO

logger = logging.getLogger(__name__)


class LinuxSCSIInterfaceException(InterfaceException):
    pass


class LinuxSCSIInterface(object):
    _port: str
    _file: BinaryIO

    def __init__(self, port: str):
        self._port = port
        try:
            # The CAP_SYS_RAWIO capability is required to communicate over SGIO
            # sudo setcap CAP_SYS_RAWIO=+ep /usr/bin/python3.8
            self._file = open(self._port, "wb")
        except (FileNotFoundError, PermissionError) as e:
            logger.error("Could not open the SCSI device: %s" % (e,))
            raise LinuxSCSIInterfaceException(e)

    def read(self, size=1):
        assert size != 0
        # Allocate data
        cdb = bytearray(0x10)
        cdb[0] = 0xEF
        cdb[1] = 0xFF  # read
        data_in = bytearray(size)
        sgio.execute(self._file, cdb, None, data_in)

        return data_in

    def write(self, data):
        # Allocate data
        cdb = bytearray(0x10)
        cdb[0] = 0xEF
        cdb[1] = 0xFE  # write
        data_out = bytearray(data)
        sgio.execute(self._file, cdb, data_out, None)

        return data_out

    def close(self):
        self._file.close()



import ctypes
import io
import struct

from collections import OrderedDict

reverse = lambda x: {v: k for k, v in x.items()}

CommandStartCode1 = ("B", (0x55,))
CommandStartCode2 = ("B", (0xAA,))

DataStartCode1 = ("B", (0x5A,))
DataStartCode2 = ("B", (0xA5,))

DeviceId = ("<H", (1,))

Parameter = lambda x: ("<L", (x,))
Command = lambda x: ("<H", (x,))
Response = lambda x: ("<H", (x,))

Checksum = lambda x: ("<H", (x,))
Data = lambda x: ("B" * len(x), (x,))

command_codes = {
    "OPEN": 0x01,
    "CLOSE": 0x02,
    "USB_INTERNAL_CHECK": 0x03,
    "CHANGE_BAUDRATE": 0x04,
    "MODULE_INFO": 0x06,
    "CMOS_LED": 0x12,
    "ENROLL_COUNT": 0x20,
    "CHECK_ENROLLED": 0x21,
    "ENROLL_START": 0x22,
    "ENROLL1": 0x23,
    "ENROLL2": 0x24,
    "ENROLL3": 0x25,
    "IS_PRESS_FINGER": 0x26,
    "DELETE_ID": 0x40,
    "DELETE_ALL": 0x41,
    "VERIFY": 0x50,
    "IDENTIFY": 0x51,
    "VERIFY_TEMPLATE": 0x52,
    "IDENTIFY_TEMPLATE": 0x53,
    "CAPTURE": 0x60,
    "MAKE_TEMPLATE": 0x61,
    "GET_IMAGE": 0x62,
    "GET_RAWIMAGE": 0x63,
    "GET_TEMPLATE": 0x70,
    "SET_TEMPLATE": 0x71,
    "GET_DATABASE_START": 0x72,
    "GET_DATABASE_END": 0x73,
    "FW_UPDATE": 0x80,
    "ISO_UPDATE": 0x81,
    "FAKE_DETECTOR": 0x91,
    "SET_SECURITY_LEVEL": 0xF0,
    "GET_SECURITY_LEVEL": 0xF1,
    "IDENTIFY_TEMPLATE_2": 0xF4,
    "STANDBY_MODE": 0xF9,
    "ACK_OK": 0x30,
    "NACK_INFO": 0x31,
}

ACK_OK = command_codes["ACK_OK"]

response_error = {
    "NACK_TIMEOUT": 0x1001,
    "NACK_INVALID_BAUDRATE": 0x1002,
    "NACK_INVALID_POS": 0x1003,
    "NACK_IS_NOT_USED": 0x1004,
    "NACK_IS_ALREADY_USED": 0x1005,
    "NACK_COMM_ERR": 0x1006,
    "NACK_VERIFY_FAILED": 0x1007,
    "NACK_IDENTIFY_FAILED": 0x1008,
    "NACK_DB_IS_FULL": 0x1009,
    "NACK_DB_IS_EMPTY": 0x100A,
    "NACK_TURN_ERR": 0x100B,
    "NACK_BAD_FINGER": 0x100C,
    "NACK_ENROLL_FAILED": 0x100D,
    "NACK_IS_NOT_SUPPORTED": 0x100E,
    "NACK_DEV_ERR": 0x100F,
    "NACK_CAPTURE_CANCELED": 0x1010,
    "NACK_INVALID_PARAM": 0x1011,
    "NACK_FINGER_IS_NOT_PRESSED": 0x1012,
}


class Packet(object):
    def __init__(self):
        self._fields = OrderedDict()
        self._fields["CommandStartCode1"] = CommandStartCode1
        self._fields["CommandStartCode2"] = CommandStartCode2
        self._fields["DeviceId"] = DeviceId

    def _checksum(self):
        return sum(self._field_bytes()) % 2 ** 16

    def byte_size(self):
        field_byte_size = sum(
            struct.calcsize(field) for field, _ in self._fields.values()
        )
        checksum_field, _ = Checksum(0)
        return field_byte_size + struct.calcsize(checksum_field)

    def _field_bytes(self):
        field_bytes = bytes()
        for _, (field, contents) in self._fields.items():
            field_bytes += struct.pack(field, *contents)
        return field_bytes

    def to_bytes(self):
        field_bytes = self._field_bytes()
        checksum_field, checksum_content = Checksum(self._checksum())

        return field_bytes + struct.pack(checksum_field, *checksum_content)

    @classmethod
    def from_bytes(cls, input_bytes):
        instance = cls()
        byte_stream = io.BytesIO(input_bytes)
        for key, (field, _) in instance._fields.items():
            field_size = struct.calcsize(field)
            field_content = byte_stream.read(field_size)
            if len(field_content) == 0:
                print("Could not parse %s" % (cls),)
                return None, None

            unpacked_value = struct.unpack(field, field_content)
            instance._fields[key] = (field, unpacked_value)

        # TODO - verify base fields

        # verify checksum
        checksum_field, _ = Checksum(0)
        checksum_bytes = byte_stream.read(struct.calcsize(checksum_field))
        if len(checksum_bytes) == 0:
            print("Checksum bytes are missing.")
            return None, None

        checksum = struct.unpack(checksum_field, checksum_bytes)[0]
        if checksum != instance._checksum():
            print("Bad checksum.")
            return None, None

        if byte_stream.tell() < len(input_bytes):
            print("Extra bytes in packet, returning in second param")

        return instance, input_bytes[byte_stream.tell() :]


class CommandPacket(Packet):
    def __init__(self, parameter=0, command=0):
        super().__init__()
        self._fields["Parameter"] = Parameter(parameter)
        self._fields["Command"] = Command(command)

    @property
    def parameter(self):
        return self._fields["Parameter"][1][0]

    @property
    def command(self):
        return self._fields["Command"][1][0]


class ResponsePacket(Packet):
    def __init__(self, parameter=0, response=0):
        super().__init__()
        self._fields["Parameter"] = Parameter(parameter)
        self._fields["Response"] = Response(response)

    @property
    def parameter(self):
        return self._fields["Parameter"][1][0]

    @property
    def response_code(self):
        return self._fields["Response"][1][0]

    @property
    def ok(self):
        return self.response_code == command_codes["ACK_OK"]


class DataPacket(Packet):
    def __init__(self, data=b""):
        self._fields = OrderedDict()
        self._fields["DataStartCode1"] = DataStartCode1
        self._fields["DataStartCode1"] = DataStartCode2
        self._fields["DeviceId"] = DeviceId

        self._fields["Data"] = ("B", data)  # abstract

    @property
    def data(self):
        return self._fields["Data"][1]


Sensor = lambda x: ("B" * 12, x)
EngineVersion = lambda x: ("B" * 12, x)
RawImgWidth = lambda x: ("<H", x)
RawImgHeight = lambda x: ("<H", x)
ImgWidth = lambda x: ("<H", x)
ImgHeight = lambda x: ("<H", x)
EnrollCount = lambda x: ("<H", x)
MaxRecordCount = lambda x: ("<H", x)
TemplateSize = lambda x: ("<H", x)


class ModuleInfoDataPacket(Packet):
    def __init__(
        self,
        sensor=b" " * 12,
        engine_version=b" " * 12,
        raw_img_width=0,
        raw_img_height=0,
        img_width=0,
        img_height=0,
        max_record_count=0,
        enroll_count=0,
        template_size=0,
    ):
        super().__init__()
        self._fields["Sensor"] = Sensor(sensor)
        self._fields["EngineVersion"] = EngineVersion(engine_version)
        self._fields["RawImgWidth"] = RawImgWidth(raw_img_width)
        self._fields["RawImgHeight"] = RawImgHeight(raw_img_height)
        self._fields["ImgWidth"] = ImgWidth(img_width)
        self._fields["ImgHeight"] = ImgHeight(img_height)
        self._fields["MaxRecordCount"] = MaxRecordCount(max_record_count)
        self._fields["EnrollCount"] = EnrollCount(enroll_count)
        self._fields["TemplateSize"] = TemplateSize(TemplateSize)

    @property
    def sensor(self):
        data = bytes(self._fields["Sensor"][1])
        return ctypes.create_string_buffer(data).value.decode("latin-1")

    @property
    def engine_version(self):
        data = bytes(self._fields["EngineVersion"][1])
        return ctypes.create_string_buffer(data).value.decode("latin-1")

    @property
    def raw_img_width(self):
        return self._fields["RawImgWidth"][1][0]

    @property
    def raw_img_height(self):
        return self._fields["RawImgHeight"][1][0]

    @property
    def img_width(self):
        return self._fields["ImgWidth"][1][0]

    @property
    def img_height(self):
        return self._fields["ImgHeight"][1][0]

    @property
    def max_record_count(self):
        return self._fields["MaxRecordCount"][1][0]

    @property
    def enroll_count(self):
        return self._fields["EnrollCount"][1][0]

    @property
    def template_size(self):
        return self._fields["TemplateSize"][1][0]


FirmwareVersion = lambda x: ("<L", x)
IsoAreaMaxSize = lambda x: ("<L", x)
DeviceSerialNumber = lambda x: ("B" * 16, x)


class OpenDataPacket(Packet):
    def __init__(
        self, firmware_version=0, iso_area_max_size=0, device_serial_number=b"0" * 16
    ):
        super().__init__()
        self._fields["FirmwareVersion"] = FirmwareVersion(firmware_version)
        self._fields["IsoAreaMaxSize"] = IsoAreaMaxSize(iso_area_max_size)
        self._fields["DeviceSerialNumber"] = DeviceSerialNumber(device_serial_number)

    @property
    def firmware_version(self):
        return hex(self._fields["FirmwareVersion"][1][0]).lstrip("0x")

    @property
    def iso_area_max_size(self):
        return self._fields["IsoAreaMaxSize"][1][0]

    @property
    def device_serial_number(self):
        return bytes(self._fields["DeviceSerialNumber"][1]).hex().upper()


Bitmap = lambda x: ("52116B", x)


class GetImageDataPacket(Packet):
    def __init__(self, bitmap=b"" * 52116):
        super().__init__()
        self._fields["Bitmap"] = Bitmap(bitmap)

    @property
    def bitmap(self):
        return bytes(self._fields["Bitmap"][1])


RawBitmap = lambda x: ("19200B", x)


class GetRawImageDataPacket(Packet):
    def __init__(self, raw_bitmap=b"" * 19200):
        super().__init__()
        self._fields["RawBitmap"] = RawBitmap(raw_bitmap)

    @property
    def raw_bitmap(self):
        return bytes(self._fields["RawBitmap"][1])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['GT521F32', 'GT521F32.interfaces', 'viewer']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=7.2.0,<8.0.0', 'cython-sgio>=1.1.2,<2.0.0', 'pyserial>=3.4,<4.0']

setup_kwargs = {
    'name': 'pygt521f32',
    'version': '0.1.0',
    'description': 'An SDK module for the ADH-Tech GT521F32 fingerprint scanner',
    'long_description': None,
    'author': 'Alon',
    'author_email': 'alonlivne@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
